import time
from influxdb import InfluxDBClient

class InfluxDBBuilder(object):
    def __init__(self, host="localhost", port=8086, precision='s', user="root", password="root"):
        self.host = host
        self.port = port
        self.precision = precision
        self.set_flag = False
        self.user = user
        self.password = password
        self.current_tag_list = {}
        self.database = "example"
        self.measurement = "measurement"
        self.client = None
    
    def reset_tags(self):
        self.current_tag_list = {}
    
    def reset(self):
        """ Reset the host for the influxdb """
        self.set_flag = True
        
        self.client = InfluxDBClient(self.host, self.port, self.user, self.password)
        self.client.create_database(self.database)
        return self
    
    

    def check_set(self):
        """Calls the reset option if it hasn't been set yet"""
        if self.set_flag == False:
            self.reset()
        self.client.switch_database(self.database)
        return self
    
    def set_user_information(self, user="root", password="root"):
        self.user = user
        self.password = password
        return self
    
    def set_db(self, name="database"):
        self.database = name
        return self
    
    def set_measurement(self, measurement="metric_name"):
        """Add a measurement name"""
        self.measurement = measurement
        return self
    
    def add_tag(self, key, value):
        self.current_tag_list[key] = value
        return self
    
    def json_builder(self, fields:dict, timestamp:float):
        return [{
            "measurement": self.measurement,
            "tags": self.current_tag_list,
            "time": int(timestamp),
            "fields": fields
        }]
    
    def write(self, fields, timestamp=None):
        if timestamp is None:
            timestamp=time.time()
        self.check_set()
        save_json = self.json_builder(fields=fields, timestamp=timestamp)
        self.client.write_points(save_json, time_precision=self.precision)
        return self
    
    def query(self, search_string):
        self.check_set()
        return self.client.query(search_string)


from .event import Event
from .hub import GenericHub
from .service import BaseService
from .pipenode import PNode
from .store_item import StoreItem
from .singleton import Singleton
from .memory import LocalMemory, RedisMemory


class StoreItem(object):
    """
        # StoreItem
        ---
        Save and load an item with dynamic programming
    """
    def __init__(self, *args, **kwargs):
        pass


    def save(self, id, obj):
        # Save the item
        raise NotImplementedError
    
    def load(self, id):
        raise NotImplementedError






import threading



class GenericHub(object):
    """ Hubs help create conditional locks. Use as a form of dynamic programming """
    def __init__(self, *args, **kwargs):
        self.action_log = {}
        self.lock = threading.Lock()

    def create(self, _id: str):
        with self.lock:
            self.action_log[_id] = []

    def push(self, _id: str, state_name: str):
        if isinstance(state_name, str):
            with self.lock:
                self.action_log[_id].append(_id)

    def get(self, _id: str):
        hub_obj = self.action_log.get(_id, None)
        if hub_obj is None:
            raise AttributeError(f"Item With Id: {_id} doesn't exist")
        
        return hub_obj
    
    def get_last(self, _id: str):
        hub_obj = self.action_log.get(_id, None)
        if hub_obj is None:
            raise AttributeError(f"Item With Id: {_id} doesn't exist")

        if len(hub_obj) == 0:
            raise IndexError(f"Item with Id: {_id}. Doesn't have any items. ")

        return hub_obj[-1]




# threading.active_count() 
# 



import uuid
import threading
from threading import Lock


class Event(object):
    def __init__(self, *args, **kwargs):
        self.type = kwargs.get("_type", "MARKET")
        self.eid = kwargs.get("eid", str(uuid.uuid4()))
        self.start_number = 1
        

    
    def set_vars(self, **kwargs):
        """
            # Set Variables
            Set important variables for the user. 
            Key Variables:
                - type: This is the main variable for events
        """
        _type = kwargs.get("_type")
        
        if _type is not None:
            self.type = _type
    
    def __setattr__(self, name, value):
        self.__dict__[name] = value



import sys
import ast
import uuid
import time
import base64
import threading
from funguauniverse import StoreItem
from loguru import logger
from spaceman import Spaceman
from hashlib import sha1
config = {
    "handlers": [
        {"sink": sys.stdout, "format": "{time:YYYY-MM-DD at HH:mm:ss} | {level} | {message}"},
    ],
    "extra": {"user": "someone"}
}
logger.configure(**config)


class MemoizeAndOperate(StoreItem, threading.Thread):
    def __init__(self, **kwargs):
        threading.Thread.__init__(self)
        self.lock = threading.Lock()
        self.interval = kwargs.get("interval", 3)
        self.mongo_host = kwargs.get("mongo_host")
        self.is_lock = kwargs.get("is_lock", True)
        self.reg_dict = {}
        self.timestamp_record = {}
        self.query_lookup_table = {}
        self.bgprocess = threading.Thread(
            target=self._run, name='bgprocess', daemon=True)
        self.bgprocess.start()
        self.space = Spaceman(mongo_host=self.mongo_host)
        self.daemon = True

    def filter_query(self, query: dict):
        assert query.get("type") is not None
        assert isinstance(query['type'], str)

    def set_item(self, query: dict, item, **kwargs):
        self.filter_query(query)
        if self.is_lock == True:
            with self.lock:
                is_overwrite = kwargs.get("overwrite", False)
                storage_string = self.hash_dict(query)
                # Add feature to check if we should overwrite
                if is_overwrite == True:
                    self.reg_dict[storage_string] = item
                    return

                if self.reg_dict.get(storage_string, None) is None:
                    self.reg_dict[storage_string] = item
        else:
            is_overwrite = kwargs.get("overwrite", False)
            storage_string = self.hash_dict(query)
            # Add feature to check if we should overwrite
            if is_overwrite == True:
                self.reg_dict[storage_string] = item
                return

            if self.reg_dict.get(storage_string, None) is None:
                self.reg_dict[storage_string] = item

    def get_item(self, query_dict: dict):
        self.filter_query(query_dict)
        if self.is_lock == True:
            with self.lock:
                storage_string = self.hash_dict(query_dict)
                item = self.reg_dict.get(storage_string, None)
                # Set that an item was touched

                self.timestamp_record[storage_string] = time.time()
                return item
        else:
            storage_string = self.hash_dict(query_dict)
            item = self.reg_dict.get(storage_string, None)
            self.timestamp_record[storage_string] = time.time()
            return item

    def hash_dict(self, query_dict: dict):
        # Updated hash for consistency between runs
        qhash = sha1(repr(sorted(query_dict.items())).encode()).hexdigest()
        self.query_lookup_table[qhash] = query_dict
        return qhash

    def query_by_hash(self, _hash):
        query_dict = self.query_lookup_table.get(_hash, None)
        return query_dict

    def dict_to_b64(self, query_dict: dict):
        qstring = str(query_dict)
        encoded = qstring.encode()
        byteb64 = base64.b64encode(encoded)
        return byteb64.decode("utf-8")

    def b64_to_dict(self, bstring: str):
        bstring_decoded = bstring.encode()
        decoded_base = base64.b64decode(bstring_decoded)
        decoded_string = decoded_base.decode()
        return ast.literal_eval(decoded_string)

    def background_operation(self):
        reg_keys = list(self.reg_dict.keys())
        for rk in reg_keys:
            latest_update = time.time()
            # self.timestamp_record[rk] = latest_update
            b64key = self.b64_to_dict(rk)
            # We would use this dict to load the most recent model.
            logger.info(f"Processing Keys: {rk}", enqueue=True)

    def _run(self):
        while True:
            self.background_operation()
            time.sleep(self.interval)

    def save(self):
        raise NotImplementedError(
            "Ensure to use this function to save the model")

    def load(self):
        raise NotImplementedError(
            "Ensure to implement this function to load the models you have")


if __name__ == "__main__":
    memop = MemoizeAndOperate(interval=0.07)
    latest_hex = uuid.uuid4().hex
    for i in range(100):
        memop.set_item({"eid": latest_hex}, 1)
        latest = memop.get_item({"eid": latest_hex})
        latest += 1
        memop.set_item({"eid": latest_hex}, latest, overwrite=True)
        print(latest)
        time.sleep(0.01)



from socketserver import ThreadingMixIn
from queue import Queue
import threading
import socket
import time


# From http://code.activestate.com/recipes/574454-thread-pool-mixin-class-for-use-with-socketservert/
# By Michael Palmer, 20 jul 2008

# This class creates a Queue where incoming requests are stored. A number of
# threads, defined by 'numThreads' then get requests off of this queue and
# handle them. The original code limited the request queue to the number of
# threads; this limitation has been removed.
class ThreadPoolMixIn(ThreadingMixIn):
    '''
    use a thread pool instead of a new thread on every request
    '''
    allow_reuse_address = True  # seems to fix socket.error on server restart
    last_time = 0
    reqs = {}

    def serve_forever(self, numThreads=20):
        '''
        Handle one request at a time until doomsday.
        '''
        self.numThreads = numThreads

        # set up the request pool
        self.requests = Queue()

        for x in range(self.numThreads):
            t = threading.Thread(target=self.process_request_thread)
            t.daemon = True
            t.start()

        # server main loop
        while True:
            self.handle_request()

        self.server_close()

    def process_request_thread(self):
        '''
        obtain request from queue instead of directly from server socket
        '''
        while True:
            ThreadingMixIn.process_request_thread(self, *self.requests.get())

    def handle_request(self):
        '''
        simply collect requests and put them on the queue for the workers.
        '''
        try:
            request, client_address = self.get_request()
        except socket.error:
            return

        if self.verify_request(request, client_address):
            self.requests.put((request, client_address))



import random
import time
import uuid
from concurrent.futures import ThreadPoolExecutor
import json
import dask.bag as db
import numpy as np
import pandas as pd



from requests_futures.sessions import FuturesSession











class PortfolioClient(object):
    """ 
        # Portfolio Client
        ---
        This is a single portfolio client. 
        We use it to test the flow through the pipeline.


        After the first trial, work to move multi-users.

        ### The Official Order
        ---
        **Create User** -> **Determine Order** -> **Send Order** -> **(Get User Stats | Get Performance)** -> **Update Context Service** -> **Get Context For Reward**
    """
    def __init__(self, host="0.0.0.0", port=9005):
        # self.session = FuturesSession(executor=ThreadPoolExecutor(max_workers=5))
        self.session = FuturesSession(ThreadPoolExecutor(max_workers=1000))
        self.address = f"http://{host}:{port}"

    def create_user(self, userid, exchange="baseexchange", is_live=False, episode=None):
        create_call = "/portfolio/create"
        call_loc = f"{self.address}{create_call}"

        call_params = {
            "userid": userid,
            "exchange": exchange,
            "spec": {
                "episodeid": episode,
                "live": is_live
            }
        }
        call_params_json = json.dumps(call_params)
        self.session.post(call_loc, data=call_params_json)



    def get_user(self, userid, exchange="baseexchange", is_live=False, episode=None):
        create_call = "/portfolio/get"
        call_loc = f"{self.address}{create_call}"
        call_params = {
            "userid": userid,
            "exchange": exchange,
            "spec": {
                "episodeid": episode,
                "live": is_live
            }
        }

        call_params_json = json.dumps(call_params)

        response = self.session.post(call_loc, data=call_params_json)
        return response.result()
    

    def get_performance(self, userid, exchange="baseexchange", is_live=False, episode=None):
        create_call = "/portfolio/performance"
        call_loc = f"{self.address}{create_call}"
        call_params = {
            "userid": userid,
            "exchange": exchange,
            "spec": {
                "episodeid": episode,
                "live": is_live
            }
        }

        call_params_json = json.dumps(call_params)

        response = self.session.post(call_loc, data=call_params_json)
        return response.result()
    

    def update_user(self, userid, exchange="baseexchange", is_live=False, episode=None, timestamp=time.time()):
        create_call = "/portfolio/update"
        call_loc = f"{self.address}{create_call}"
        call_params = {
            "userid": userid,
            "exchange": exchange,
            "spec": {
                "episodeid": episode,
                "live": is_live
            },
            "timestamp": timestamp
        }

        call_params_json = json.dumps(call_params)

        response = self.session.post(call_loc, data=call_params_json)
        return response
        


    def get_user_stats(self, userid, exchange="baseexchange", is_live=False, episode=None):
        create_call = "/portfolio/stats"
        call_loc = f"{self.address}{create_call}"
        call_params = {
            "userid": userid,
            "exchange": exchange,
            "spec": {
                "episodeid": episode,
                "live": is_live
            }
        }
        call_params_json = json.dumps(call_params)

        response = self.session.post(call_loc, data=call_params_json)
        return response.result()
    

    def send_order(self, 
            userid:int, percentage:float, order_type:str,
            base:str, trade:str, price:float, exchange="baseexchange", 
            is_live=False, episode=None, timestamp=time.time()):
        """
            print(userid)
            print(percentage)
            print(order_type)
            print(base)
            print(trade)
            print(price)
            print(exchange)
            print(is_live)
            print(episode)
            print(timestamp)
        """
        create_call = "/order/send"
        call_loc = f"{self.address}{create_call}"
        call_params = {
            "userid": userid,
            "base": base,
            "trade": trade,
            "exchange": exchange,
            "pct": percentage,
            "action": order_type,
            "price": price,
            "spec": {
                "episodeid": episode,
                "live": is_live
            },
            "timestamp": timestamp
        }
        call_params_json = json.dumps(call_params)
        self.session.post(call_loc, data=call_params_json)

    def user_allocation(self, userid, assets, exchange="baseexchange", is_live=False, episode=None):
        
        create_call = "/portfolio/allocate"
        call_loc = f"{self.address}{create_call}"
        call_params = {
            "userid": userid,
            "exchange": exchange,
            "spec": {
                "episodeid": episode,
                "live": is_live
            },
            "assets": assets
        }
        call_params_json = json.dumps(call_params)
        self.session.post(call_loc, data=call_params_json)
    # TODO: Get coin status for a given user
    # We can use that for 5-10 users concurrently
    # The switch from there will be integrating with the existing calls we have
    # Test the environment with one user first. Prepare to add other users into the mix as well
    # After you get the portoflio and context manager worked out, move forward

if __name__ == "__main__":
    # Test Calls
    episode_eid = uuid.uuid4().hex
    userid = 6000
    portfolio_client = PortfolioClient()
    for _ in range(1000):
        portfolio_client.create_user(_, is_live=False, episode=episode_eid)
        # user = portfolio_client.get_user(_, is_live=False, episode=episode_eid)
        # performance = portfolio_client.get_performance(_, is_live=False, episode=episode_eid)
        order_type = random.choice(["BUY", "HOLD", "SELL"])
        random_price = random.uniform(100, 1000)
        portfolio_client.send_order(
            userid=_,
            percentage=random.uniform(0, 1),
            order_type=order_type,
            base="USD",
            trade="BTC",
            price=random_price,
            exchange="backtest",
            is_live=False,
            episode=episode_eid
        )
        
        # stats = portfolio_client.get_user_stats(userid, is_live=False, episode=episode_eid)
        # print(stats.json())
    # print(user.json())
    # print(performance.json())
    # print("----------")
    # print("----------")
    # print(stats.json())
    
    



import requests
from concurrent.futures import ProcessPoolExecutor
from requests_futures.sessions import FuturesSession
import pickle
from loguru import logger
from requests import Session



class ServiceClient(object):
    """ Call the HTTP Service directly """
    def __init__(self, **kwargs):
        self.host = kwargs.get("address", "localhost")
        self.port = kwargs.get("port", 5581)
        self._address = f"http://{self.host}:{self.port}"
        self.session = FuturesSession(max_workers=1000)

    def _send(self, data, blocking=True):
        payload = pickle.dumps(data)
        
        response = self.session.post(self._address, data=payload)
        if blocking == True:
            response = response.result()
            if response.status_code != 200:
                logger.error("Request failed {}: {}".format(response.text, data))
            try:
                response.raise_for_status()
                parsed = pickle.loads(response.content)
                return parsed
            except:
                return {}
        else:
            logger.info("async-call", enqueue=True)
        


if __name__ == "__main__":
    serve_command = ServiceClient()



import uuid
import requests
import pandas as pd
import sys
from funguauniverse import ServiceClient
from loguru import logger

config = {
    "handlers": [
        {"sink": sys.stdout, "format": "{time} - {message}"},
        {"sink": "file.log", "serialize": True},
    ],
    "extra": {"user": "someone"}
}
logger.configure(**config)


class PreprocessingClient(ServiceClient):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def ta_prices(self, episode_id, coin, price:pd.DataFrame):
        price_dict = price.to_json(orient='index')
        response = self._gsend("/preprocess/prices", {
            "episode": episode_id,
            "coin": coin,
            "data": price_dict
        })
        return response

    def _gsend(self, route, data):
        """ General Send """
        address = self._address+route
        response = requests.post(address, json=data)
        if response.status_code != 200:
            logger.error("Request failed {}".format(response.text))
        parsed = response.content
        return parsed

if __name__ == "__main__":
    pass
    # assert action.get("pct", None) is not None
    # assert action.get("price", None) is not None
    # assert action.get("coin", None) is not None
    # assert action.get("decision", None) is not None
    # import random
    # import time
    # current_eid = '1dd20719728b49148441bb23c0ff84d7'
    # client = ContextualAwareness(host="localhost", port=9847)
    # client.initialize(current_eid)
    # action = {"pct": 0.2, "price": 270 *
    #           random.normalvariate(0.5, 0.2), "coin": "ETH", "decision": 0.5}
    # timestamp = time.time()
    # client.log_action(current_eid, action, timestamp)

    # start = time.time()
    # index = 0

    # for i in range(1000):
    #     client.log_reward(current_eid, "portfolio", random.normalvariate(
    #         12000, 2000), time.time() + (index * 60))
    # end = time.time()

    # print(end-start)

# eyd0eXBlJzogJ2NvbnRleHQnLCAnZWlkJzogJzA3MjA3M2UxNTcyYzRlOTViNWQ5NTY4MjAxN2NlZjc5J30=
# eyd0eXBlJzogJ2NvbnRleHQnLCAnZWlkJzogJ2MyMzVjZjRlMmIwYjQxZGQ4YjE5ZDBiMDIyMTZkMzY5J30=



import json
import time
from concurrent.futures import ThreadPoolExecutor
from requests_futures.sessions import FuturesSession
from loguru import logger



class MemoryService(object):
    def __init__(self, host="0.0.0.0", port=8500, is_ssl=False):
        protocol = "http"
        if is_ssl == True:
            protocol = "https"
        self.call_variable = f"{protocol}://{host}:{port}"
        self.session = FuturesSession(executor=ThreadPoolExecutor(max_workers=100))

    def store_action(self, percentage, decision, price, pair, episode=None, setid=None, is_live=False):
        """ Returns the latest price for a pair"""
        
        logger.info(f"Storing action for coin: {pair}")
        try:
            return self._send({
                "call": "basic/store/actions",
                "data": {
                    "percentage": percentage,
                    "decision": decision,  # Will try grabbing from live if it's None
                    "price": price,
                    "spec": {
                        "pair": pair,
                        "live": is_live,
                        "episode": episode,
                        "setid": setid
                    }
                    
                }
            })
        except Exception:
            logger.error("Call Failed - get_single")
            return {
                "data": {}
            }

    def store_reward(self, reward, entity, pair, episode=None, setid=None, is_live=False):
        """ Returns the latest price for a pair"""
        logger.info(f"Storing reward for coin: {pair} -- with entity: {entity}")
        try:
            return self._send({
                "call": "basic/store/rewards",
                "data": {
                    "reward": reward,
                    "entity": entity,
                    "spec": {
                        "pair": pair,
                        "live": is_live,
                        "episode": episode,
                        "setid": setid
                    }
                }
            })
        except Exception:
            logger.error("Call Failed - get_single")
            return {
                "data": {}
            }
    

    def query_reward(self, entity, pair, episode=None, setid=None, is_live=False):
        """ Returns the latest price for a pair"""
        print("Get the single latest price")
        try:
            return self._send({
                "call": "basic/query/rewards",
                "data": {
                    "entity": entity,
                    "spec": {
                        "pair": pair,
                        "live": is_live,
                        "episode": episode,
                        "setid": setid
                    }
                    
                }
            })
        except Exception:
            logger.error("Call Failed - get_single")
            return {
                "data": {}
            }


    def query_action(self, pair, episode=None, setid=None, is_live=False):
        """ Returns the latest price for a pair"""
        print("Get the single latest price")
        try:
            return self._send({
                "call": "basic/query/actions",
                "data": {
                    "spec": {
                        "pair": pair,
                        "live": is_live,
                        "setid": setid,
                        "episode": episode
                    }
                    
                }
            })
        except Exception:
            logger.error("Call Failed - get_single")
            return {
                "data": {}
            }


    def log_counterfactual(self, covariate, instrument, treatment, pair, bartime=time.time(), episode=None, setid=None, is_live=False):
        counterfactual_json = {
            "covariance": covariate,
            "instruments": instrument,
            "treatment": treatment,
            "bartime": bartime,
            "spec": {
                "pair": pair,
                "live": is_live,
                "setid": setid,
                "episode": episode
            }
        }

        try:
            return self._send({
                "call": "counterfact/store",
                "data": counterfactual_json
            })
        except Exception:
            logger.error("Call Failed - get_single")
            return {
                "data": {}
            }

    def log_outcome(self, outcome, entity, pair, bartime=time.time(), episode=None, setid=None, is_live=False):
        counterfactual_json = {
            "outcome": outcome,
            "entity": entity,
            "bartime": bartime,
            "spec": {
                "pair": pair,
                "live": is_live,
                "setid": setid,
                "episode": episode
            }
        }

        try:
            return self._send({
                "call": "counterfact/store/outcome",
                "data": counterfactual_json
            })
        except Exception:
            logger.error("Call Failed - get_single")
            return {
                "data": {}
            }
    
    def query_outcome(self, entity, pair, max_num=50, episode=None, setid=None, is_live=False):
        outcome_json = {
            "entity": entity,
            "max_num": max_num,
            "spec": {
                "pair": pair,
                "live": is_live,
                "setid": setid,
                "episode": episode
            }
        }

        try:
            return self._send({
                "call": "counterfact/query/all",
                "data": outcome_json
            })
        except Exception:
            logger.error("Call Failed - get_single")
            return {
                "data": {}
            }

    def _send(self, call_info: dict):
        call = call_info.get("call")
        data = call_info.get("data")
        assert call is not None
        assert data is not None

        concatenated_call = self.call_variable + f"/{call}"
        
        resp = self.session.post(concatenated_call, json=data)
        return resp


import json
from loguru import logger

from concurrent.futures import ThreadPoolExecutor
from requests_futures.sessions import FuturesSession



class PriceService(object):
    def __init__(self, host="0.0.0.0", port=8500):
        self.call_variable = f"http://{host}:{port}"
        self.session = FuturesSession(executor=ThreadPoolExecutor(max_workers=5))

    def get_single(self, pair, episode=None, is_live=False):
        """ Returns the latest price for a pair"""
        print("Get the single latest price")
        try:
            return self._send({
                "call": "price/get_single",
                "data": {
                    "coinname": pair,
                    "episode": episode,  # Will try grabbing from live if it's None
                    "limit": 250,
                    "is_live": is_live
                }
            })
        except Exception:
            logger.error("Call Failed - get_single")
            return {
                "data": {}
            }

    def get_single_before(self, pair, episode=None, is_live=False):
        """ Returns the latest price for a pair"""
        print("Get the latest prices")
        try:
            return self._send({
                "call": "price/get_single_before",
                "data": {
                    "coinname": pair,
                    "episode": episode,  # Will try grabbing from live if it's None
                    "limit": 250,
                    "is_live": is_live
                }
            })
        except Exception:
            logger.error("Call Failed - get_single_before")
            return {
                "data": {}
            }

    def get_multi(self, pair, episode=None, is_live=False):
        # get_multi
        try:
            # Will try grabbing from live if it's None
            return self._send({
                "call": "price/get_multi",
                "data": {
                    "coinname": pair,
                    "episode": episode,
                    "limit": 250,
                    "is_live": is_live
                }
            })
            
        except Exception:
            logger.error("Call Failed - get_multiple_prices")
            return {
                "data": {}
            }

    def add_single(self, pair, price_bar, episode=None, is_live=False):
        try:
            return self._send({
                "call": "price/set_single",
                "data": {
                    "single": {
                        "coinname": pair,
                        "bar": price_bar
                    },
                    "episode": episode,  # Will try grabbing from live if it's None
                    "is_live": is_live
                }
            })
        except Exception:
            logger.error("Call Failed - multiple_prices")
            return {
                "data": {}
            }

    def add_multi(self, pair, price_bars, episode=None, is_live=False):
        try:
            return self._send({
                "call": "price/set_multi",
                "data": {
                    "multi": {
                        "coinname": pair,
                        "bars": price_bars
                    },
                    "episode": episode,  # Will try grabbing from live if it's None
                    "is_live": is_live
                }
            })
        except Exception:
            logger.error("Call Failed - add_multiple_prices")
            return {
                "data": {}
            }

    def _send(self, call_info: dict):
        call = call_info.get("call")
        data = call_info.get("data")
        assert call is not None
        assert data is not None

        concatenated_call = self.call_variable + f"/{call}"
        call_params_json = json.dumps(data)
        resp = self.session.post(concatenated_call, json=call_params_json)
        return resp



import json
import random
import threading
import time
from multiprocessing import Queue


from loguru import logger
import pandas as pd
import numpy as np

"""
    Acceptable Log Type
"""

class BaseService(object):
    """ 
        # Base Service 
        ---
        Here we run a base service. All services that inherit the class should work on the following methods to make it individualized:
            - `process`: Any
                - the `process` method takes the latest item added to the queue and does work on it. 
                This method is empty. Generally, you'd want to do all of the heavy lifting here.

            - `process_condition`: bool
                - The `process_condition` checks to see if a given condition has been met before processing the given variable. 
                This is good for the times when theres dependencies on previous processes that need to be syncronized
    """
    def __init__(self, result_queue, *args, **kwargs):
        self.main_queue = Queue()
        self.result_queue = result_queue
        
        # OPTIONAL VARIABLES (through kwargs)
        self.interval = kwargs.get("interval", 0.005)
        self.is_event_queue = kwargs.get("is_event_queue", False)
        self.event_queue = kwargs.get("event_queue", None)
        self.next_event_type = kwargs.get("next_ev_type", None)
        is_immediate = kwargs.get("is_immediate", True)

        
        self.thread = threading.Thread(target=self.run, args=())
        self.thread.daemon = True                            # Daemonize thread
        
        
        if is_immediate:
            self.thread.start()  # Start the execution

    

    def set_settings(self, **kwargs):
        """
            # Change Settings
            ---
            Optionally have the opprotunity to change any of the required variables for the class
        """
        _event_queue = kwargs.get("event_queue", None)
        _is_event_queue = kwargs.get("is_event_queue", None)
        _interval = kwargs.get("interval", None)
        _next_event_type = kwargs.get("next_ev_type", None)

        if self.crt(_event_queue, [Queue]):
            self.event_queue = _event_queue
        
        if self.crt(_is_event_queue, [bool]):
            self.is_event_queue = _event_queue
        
        if self.crt(_interval, [int, float]):
            self.interval = _interval
        
        if self.crt(_next_event_type, [str]):
            self.next_event_type = _next_event_type





    def push(self, item):
        """
            Push an item with the acceptable types into the main queue to be processed.

            Most specifically, we should have a blacklist of types to add
        """
        if item is None:
            logger.debug("We don't accept None types inside of the queue. ")
        
        self.main_queue.put(item)
        
    def process(self):
        """ 
            # Generic Processing
            ---
            It pulls from the main queue then processes everything that we'll need inside of the given activity.
            Keep the components of processing.

            Ensure to overload this with each run.
        """
        
        # logger.info("SUCK DICK BITCH")
        # print("Is it working?")

    def process_condititon(self):
        return True

    def run(self):
        while True:
            # item = self.__get_latest_from_queue()
            self.process()
            # logger.error(item)
            time.sleep(self.interval)

    def start(self):
        """
            Start thread if it's not immediate
        """
        if not self.thread.isAlive():
            self.thread.start()

    def get_latest_from_queue(self):
        """ Get the latest item from the message queue """
        is_empty = self.main_queue.empty()
        if not is_empty:
            __i = self.main_queue.get()
            # logger.error(__i)
            return __i
        return None

    def __check_none(self, item):
        if item is None:
            return False
        return True

    def __check_event_queue(self):
        if (self.is_event_queue == True) and (self.event_queue is None):
            return False
        elif (self.is_event_queue == True) and (self.event_queue is not None):
            return True
        elif (self.is_event_queue == False):
            return True
    
    def crt(self, item, types_list: list):
        """
            # Check Required Types
            ---
            Check the required types for the given variable

        """
        if item is None:
            return False

        item_type = type(item)
        if item_type in types_list:
            return True
        return False



from abc import ABC
import redis
import ray
# from redis import Redis

class MemoryInterface(ABC):
    def save(self, k, v, overwrite=False):
        raise NotImplementedError("You need to have a way to save information")
    
    def load(self, k):
        raise NotImplementedError("You need to have a way to load information")
    
    def remove(self, k):
        raise NotImplementedError("You need to remove keys")

    def get_keys(self):
        raise NotImplementedError("get_keys must be implemented")

class LocalMemory(MemoryInterface):
    def __init__(self, *args, **kwargs):
        self.local_dict = {}

    def save(self, k, v, overwrite=False):
        self.local_dict[k] = v
    
    def load(self, k):
        if k in self.local_dict:
            return self.local_dict[k]
        raise AttributeError(f"The key {k} doesn't exist")
    
    def remove(self, k):
        if k in self.local_dict:
            del self.local_dict[k]
        raise AttributeError(f"The key {k} doesn't exist")

    def get_keys(self):
        return list(self.local_dict.keys())

class RedisMemory(MemoryInterface):
    def __init__(self, redis_instance:redis.Redis, list_key_name="key_lists", *args, **kwargs):
        self.redis_instance = redis_instance
        self.key_list_name = list_key_name
    
    def save(self, key, val, overwrite=True):
        is_exist = self.redis_instance.get(key)
        if is_exist is not None and overwrite == False:
            return 0
        
        self.redis_instance.set(key, val)
        self.redis_instance.sadd(self.key_list_name, key)
        return 1
    def load(self, key):
        item = self.redis_instance.get(key)
        
        if item is None:
            raise AttributeError("The item does not exist in redis")
        return item
    
    def remove(self, key):
        self.redis_instance.delete(key)
        self.redis_instance.srem(self.key_list_name, key)

    def get_keys(self):
        keys = self.redis_instance.smembers(self.key_list_name)
        if keys is None:
            return []
        return [str(x) for x in keys]

class RemoteMemory(MemoryInterface):
    def __init__(self, remote_obj, *args, **kwargs):
        # The ray remote object should be a memory type. 
        self.ray_remote = remote_obj.remote()
    
    def save(self, k, v, overwrite=False):
        self.ray_remote.save.remote(k, v, overwrite)
    
    def load(self, key):
        return ray.get(self.ray_remote.load.remote(key))

    def remove(self, key):
        self.ray_remote.remove.remote(key)

    def get_keys(self):
        return ray.get(self.ray_remote.get_keys.remote())


if __name__ == "__main__":
    pass


from .utils.pipeline.event import Event
from .utils.pipeline.hub import GenericHub
from .utils.pipeline.pipenode import PNode
from .utils.pipeline.service import BaseService
from .utils.pipeline.store_item import StoreItem
from .utils.pipeline.singleton import Singleton
from .utils.pipeline.bgevent import MemoizeAndOperate
from .utils.pipeline.memory import (MemoryInterface, LocalMemory, 
                                    RedisMemory, RemoteMemory)

from .utils.web.serviceclient import ServiceClient
from .utils.web.background_service import start_service


import os
import codecs
import logging
import sys
from shutil import rmtree
from setuptools import setup, find_packages, Command
from setuptools.command.test import test as TestCommand

here = os.path.abspath(os.path.dirname(__file__))


with open("README.md", "r") as fh:
    long_description = fh.read()


class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = []

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        logging.basicConfig(
            format='%(asctime)s %(levelname)s %(name)s %(message)s', level='DEBUG')

        # import here, cause outside the eggs aren't loaded
        import pytest
        import six

        args = [self.pytest_args] if isinstance(
            self.pytest_args, six.string_types) else list(self.pytest_args)
        # args.extend(['--cov', 'arctic',
        #              '--cov-report', 'xml',
        #              '--cov-report', 'html',
        #              '--junitxml', 'junit.xml',
        #              ])
        errno = pytest.main(args)
        sys.exit(errno)


class UploadCommand(Command):
    """Support setup.py publish."""

    description = "Build and publish the package."
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print("\033[1m{0}\033[0m".format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status("Removing previous builds...")
            rmtree(os.path.join(here, "dist"))
        except FileNotFoundError:
            pass
        self.status("Building Source distribution...")
        os.system("{0} setup.py sdist bdist_wheel".format(sys.executable))
        self.status("Uploading the package to PyPi via Twine...")
        os.system("sudo twine upload dist/*")
        sys.exit()



setup(
    name="funguauniverse",
    version="0.4.2",
    author="Kevin Hill",
    author_email="kah.kevin.hill@gmail.com",
    description="A set of class tools to run the Funguana Pipeline. Many irrelelavent parts",
    long_description=long_description,
    long_description_content_type="text/markdown",
    py_modules=["funguauniverse"],
    install_requires=['pytest', 'scipy', 'numpy', 'pandas', 'dask[complete]',
                      'ta-lib', 'loguru', 'spaceman', 
                      'prometheus-client', 'requests-futures', 'influxdb'
                      ],
    packages=find_packages(exclude=['scripts']),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    cmdclass={"upload": UploadCommand, "test": PyTest},
)



def Singleton(class_):
    class class_w(class_):
        _instance = None

        def __new__(class_, *args, **kwargs):
            if class_w._instance is None:
                class_w._instance = super(class_w,
                                          class_).__new__(class_,
                                                          *args,
                                                          **kwargs)
                class_w._instance._sealed = False
            return class_w._instance

        def __init__(self, *args, **kwargs):
            if self._sealed:
                return
            super(class_w, self).__init__(*args, **kwargs)
            self._sealed = True
    class_w.__name__ = class_.__name__
    return class_w


from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import pickle
import traceback

from http.server import SimpleHTTPRequestHandler, HTTPServer
from socketserver import ThreadingMixIn
from funguauniverse.utils.web.threadpoolmixin import ThreadPoolMixIn


class ThreadedService(ThreadPoolMixIn, HTTPServer):
    def __init__(self, fun_service, address, port):
        # Create a service that
        handler = _make_handler(fun_service)
        HTTPServer.__init__(self, (address, port), handler)


def _make_handler(external_env):
    class Handler(SimpleHTTPRequestHandler):
        def do_POST(self):
            content_len = int(self.headers.get("Content-Length"), 0)
            raw_body = self.rfile.read(content_len)
            parsed_input = pickle.loads(raw_body)
            try:
                response = self.execute_command(parsed_input)
                self.send_response(200)
                self.end_headers()
                self.wfile.write(pickle.dumps(response))
            except Exception:
                self.send_error(500, traceback.format_exc())

        def execute_command(self, args):
            command = args["command"]
            arguments = args["args"]
            keywords = args["kwargs"]
            response = {}

            # print(args)
            # Create a method that allows us to determine executed command and associated action. 
            # Perhaps create a parser for this 
            if hasattr(external_env, command):
                command_resposne = getattr(external_env, command)(*arguments, **keywords)
                if command_resposne is not None:
                    response['data'] = command_resposne
            else:
                raise Exception("Unknown command: {}".format(command))
            return response

    return Handler



def start_service(env, host, port):
    service = ThreadedService(env, host, port)
    print(f"Serving Service on Port: {host}:{port}")
    service.serve_forever()


